use btlib::{
    self,
    crypto::{file_cred_store::FileCredStore, tpm::TpmCredStore, CredStore, CredStoreMut, Creds},
    error::BtErr,
    Result,
};
use config::{
    builder::{AsyncState, BuilderState, DefaultState},
    ConfigBuilder, Environment,
};
use serde::Deserialize;
use std::{path::PathBuf, sync::Arc};

/// Attempts to unwrap the [Option] field in `$config` called `$name`. This macro internal
/// uses the `?` operator, and can only be used in methods which return a [Result] which has an
/// error type that a [btlib::Error] can be converted to.
#[macro_export]
macro_rules! get_setting {
    ($config:expr, $name:ident) => {
        $config
            .$name
            .ok_or_else(|| bterr!(concat!("setting '", stringify!($name), "' was not set")))?
    };
}

/// Attempts to unwrap all of the [Option] fields with the given names in the configuration struct
/// and returns them all as a tuple. The same consideration which apply to [get_setting!] apply to
/// this macro.
#[macro_export]
macro_rules! get_settings {
    ($config:expr$(, $names:ident)+) => {
        ($(get_setting!($config, $names),)+)
    };
}

fn environment_source() -> Environment {
    Environment::with_prefix("BT").separator("_")
}

fn set_defaults<St: BuilderState>(builder: ConfigBuilder<St>) -> Result<ConfigBuilder<St>> {
    builder
        .set_default(CredStoreCfg::TAG, "File")?
        .set_default(CredStoreCfg::PATH, "./state/file_cred_store")
        .bterr()
}

/// An extension to [ConfigBuilder] which allows it to be configured using Blocktree's default
/// settings.
pub trait ConfigBuilderExt: Sized {
    /// Configure this [ConfigBuilder] with Blocktree's default settings.
    fn btconfig(self) -> Result<Self>;
}

impl ConfigBuilderExt for ConfigBuilder<DefaultState> {
    fn btconfig(self) -> Result<Self> {
        let builder = set_defaults(self)?.add_source(environment_source());
        Ok(builder)
    }
}

impl ConfigBuilderExt for ConfigBuilder<AsyncState> {
    fn btconfig(self) -> Result<Self> {
        let builder = set_defaults(self)?.add_source(environment_source());
        Ok(builder)
    }
}

#[macro_export]
macro_rules! config_keys {
    {
        const $struct_key_name:ident = $struct_key:literal;
        $(const $field_key_name:ident = $field_key:literal;)+
    } => {
        #[allow(dead_code)]
        const $struct_key_name: &str = $struct_key;
        $(
            #[allow(dead_code)]
            const $field_key_name: &str = concat!($struct_key, ".", $field_key);
        )+
    }
}

#[derive(Debug, Clone, Deserialize)]
#[serde(tag = "type")]
/// Configuration which specifies how to construct a [btlib::crypto::CredStore].
pub enum CredStoreCfg {
    /// Specifies that a [btlib::crypto::file_cred_store::FileCredStore] should be used.
    File {
        /// The path where the credential store is located.
        path: PathBuf,
    },
    /// Specifies that a [btlib::crypto::tpm::TpmCredStore] should be used.
    Tpm {
        /// The state path to use for the TPM credential store.
        path: PathBuf,
        /// The configuration string to pass to `tss-esapi`.
        tabrmd: String,
    },
}

impl CredStoreCfg {
    config_keys! {
        const STRUCT_KEY = "credstore";
        const TAG = "type";
        const PATH = "path";
        const TABRMD = "tabrmd";
    }
}

/// Trait for types which can consume a [CredStore] implementation.
pub trait CredStoreConsumer {
    type Output;
    fn consume<C: CredStore>(self, cred_store: C) -> Self::Output;
}

/// Trait for types which can consume a [CredStoreMut] implementation.
pub trait CredStoreMutConsumer {
    type Output;
    fn consume_mut<C: CredStoreMut>(self, cred_store: C) -> Self::Output;
}

macro_rules! impl_consume {
    ($name:ident, $consumer:path) => {
        #[doc = concat!("Calls `", stringify!($name), "` with the credential store created based on the configuration data in `self`")]
        pub fn $name<F: $consumer>(self, consumer: F) -> Result<F::Output> {
            match self {
                CredStoreCfg::File { path } => {
                    let store = FileCredStore::new(path)?;
                    Ok(consumer.$name(store))
                }
                CredStoreCfg::Tpm { path, tabrmd } => {
                    let store = TpmCredStore::from_tabrmd(&tabrmd, path)?;
                    Ok(consumer.$name(store))
                }
            }
        }
    }
}

impl CredStoreCfg {
    impl_consume!(consume, CredStoreConsumer);
    impl_consume!(consume_mut, CredStoreMutConsumer);
}

/// A [CredStoreConsumer] which gets the node creds from the [CredStore], then returns them
/// as an `Arc<dyn Creds>`.
pub struct NodeCredConsumer;

impl CredStoreConsumer for NodeCredConsumer {
    type Output = Result<Arc<dyn Creds>>;
    fn consume<C: CredStore>(self, cred_store: C) -> Self::Output {
        Ok(Arc::new(cred_store.node_creds()?))
    }
}
